************************************************************************
      PROGRAM smart
* V1.2.3 - 28/12/2020
* The first lines of the potential should be:
*      FUNCTION pot(t,x,n) <-- n = dimension of the phase space
*      INTEGER n
*      DOUBLE PRECISION pot,t,x(n)
* All (real) variables should be DOUBLE PRECISION.
* Names, including their dimensions, should be < 40 characters long
* Names should not begin with the letter 'o', 'O'
* No DIMENSION nor IMPLICIT sentences.
* IFs always with THEN
* No statement function statements.
* The dependent variable x should not appear with a variable index 
*    (e.g. x(k)), but always with an explicit number (e.g. x(1)).
* Auxiliary subprograms should not use derivable variables
* The following intrinsics are not expected to be present because the
*    real variables should be DOUBLE PRECISION:
*    REAL, FLOAT, SNGL, DBLE, CMPLX, DPROD, AIMAG, CONJG.
* The following intrinsics are not expected to be present because they
*    are not derivable: truncations (INT, IFIX, IDINT, AINT, DINT,
*    ANINT, DNINT, NINT, IDNINT), transfer of sign (SIGN, ISIGN,
*    DSIGN), positive difference (DIM, IDIM, DDIM), remaindering (MOD,
*    AMOD, DMOD), and maxima/minima (MAX, MAX0, AMAX1, DMAX1, AMAX0,
*    MAX1, MIN0, AMIN1, DMIN1, AMIN0, MIN1)
*
* All accelerations should be gradients of the potential (e.g. 
*    Coriolis acceleration doesn't work, unless you insert it by hand
*    in the *.acc and *.var files).
************************************************************************
      INCLUDE 'smart.par'
      INTEGER i,j,code,val,pila(piladep),npi,pastot
      CHARACTER arch*30
      LOGICAL alg,algm,chau

      INTEGER pas,veloc
      COMMON /pasada/pas,veloc

* Algebraic operations
      algm(i)=i.EQ.mul.OR.i.EQ.div.OR.i.EQ.pow ! * / **
      alg(i)=i.EQ.add.OR.i.EQ.sub.OR.algm(i) ! all
* Read potential and open files
      OPEN(50,FILE='smart.in')
      READ(50,*)arch
      READ(50,*)pastot
      CLOSE(50)
* Start cycle: 1 = accelerations; 2 = variational equations
      veloc=0 ! veloc=1: derive wrt v if the potential depends on them
      pas=1
1     chau=.FALSE.
      j=INDEX(arch,'.')
      IF(pas.EQ.1)THEN
         OPEN(50,FILE=arch,STATUS='OLD')
         OPEN(60,FILE=arch(1:j)//'acc')
      ELSE
         OPEN(50,FILE=arch(1:j)//'acc')
         OPEN(60,FILE=arch(1:j)//'var')
      ENDIF
* Find variables of the file and write header
      CALL encabezado()
* Look for the first line to be analised
      CALL copiar(chau)
* Loop over lines
      DOWHILE(.NOT.chau)
         npi=0  ! order number on encode stack
         CALL lexi(code,val)
         DOWHILE(code.NE.nline)
            IF(code.EQ.assign.OR.code.EQ.pow.OR.code.EQ.left.OR.
     &       code.EQ.coma)THEN
               npi=npi+1
               IF(npi.GT.piladep)STOP 'increase piladep in smart.par'
               pila(npi)=code
            ELSEIF(code.EQ.label)THEN
               WRITE(60,'(i5,a)')val,' CONTINUE'
            ELSEIF(code.EQ.inum.OR.code.EQ.rnum.OR.code.EQ.pvar.OR.
     &       code.EQ.rvar.OR.code.EQ.intr)THEN
               npi=npi+2
               pila(npi-1)=code
               pila(npi)=val
            ELSEIF(code.EQ.add.OR.code.EQ.sub)THEN
               DOWHILE(alg(pila(npi-2))) ! pending operations
                  CALL cierra(npi,pila,svar)
               ENDDO
               npi=npi+1
               pila(npi)=code
            ELSEIF(code.EQ.mul.OR.code.EQ.div)THEN
               DOWHILE(algm(pila(npi-2))) ! pending operations
                  CALL cierra(npi,pila,svar)
               ENDDO
               npi=npi+1
               pila(npi)=code
            ELSEIF(code.EQ.right)THEN ! right bracket
               DOWHILE(alg(pila(npi-2))) ! pending operations
                  CALL cierra(npi,pila,svar)
               ENDDO
               IF(pila(npi-4).EQ.intr.OR.pila(npi-2).EQ.coma)THEN
* Intrinsic (1 or 2 arguments) before bracket
                  CALL cierra(npi,pila,svar)
               ELSE ! delete left bracket
                  pila(npi-2)=pila(npi-1)
                  pila(npi-1)=pila(npi)
                  npi=npi-1
               ENDIF
            ENDIF
* Get new token
            CALL lexi(code,val)
         ENDDO
* New line: close pending operations
         IF(pila(npi-2).EQ.assign)THEN ! code a direct assignment
            npi=npi+3
            pila(npi-2)=noop
         ENDIF
         DOWHILE(pila(npi-5).NE.assign) ! pending operations
            CALL cierra(npi,pila,svar)
         ENDDO
         IF(pila(npi-7).EQ.rvar)THEN
            CALL cierra(npi,pila,pila(npi-7)) ! last assignment
         ELSE
            PRINT *,pila(npi-7)
            STOP 'unknown type'
         ENDIF
         CALL copiar(chau) ! search for a new line to analise
         IF(chau)pas=pas+1
      ENDDO
      IF(pas.LE.pastot)GOTO 1
      END

************************************************************************
      SUBROUTINE lexi(code,val)
      INTEGER code,val
* Get the code of the next token of the line
* Also store constants
* code: code of the token
* val: order number for constants and variables (ninum, nrnum, nrvar,
*      npvar); # intrinsic; label # for label; 0 otherwise
************************************************************************
      INCLUDE 'smart.par'
      INCLUDE 'smart.com'

      INTEGER nrvar,npvar
      COMMON /fijos/nrvar,npvar ! shared with encabezado & copiar

      CHARACTER intrin(maxstd)*5
      DATA intrin/'ABS','SQRT','EXP','LOG','LOG10','SIN','COS','TAN',
     & 'ASIN','ACOS','ATAN','SINH','COSH','TANH','ATAN2'/
      CHARACTER specific(24)*6
      DATA specific/'IABS','DABS','CABS','DSQRT','CSQRT','DEXP','CEXP',
     & 'ALOG','DLOG','CLOG','ALOG10','DLOG10','DSIN','CSIN','DCOS',
     & 'CCOS','DTAN','DASIN','DACOS','DATAN','DSINH','DCOSH','DTANH',
     & 'DATAN2'/

      DOUBLE PRECISION ator
      CHARACTER c*1,cad*40,d*1,ma*1369,rowb*1369
      INTEGER lpos,i,ninum,nrnum,eti,cl
      LOGICAL letra,digito,nume,pend
      DATA pend,ninum,nrnum,lpos/.FALSE.,3*0/
      SAVE pend,ninum,nrnum,lpos,rowb

      letra(c)=('A'.LE.c.AND.c.LE.'Z').OR.('a'.LE.c.AND.c.LE.'z')
      digito(c)='0'.LE.c.AND.c.LE.'9'

      IF(pend)THEN ! send '*' after an unary minus
         pend=.FALSE.
         code=mul
         RETURN
      ENDIF
      IF(lpos.EQ.0.AND.code.NE.label)THEN ! read a new line
         CALL readl(1,rowb,eti,cl)
         IF(eti.NE.0)THEN
            code=label
            RETURN ! next time continue from below
         ENDIF
      ENDIF
      cad=' '
      val=0
      nume=.FALSE.
      lpos=lpos+1
* Get the next character of the line
      c=rowb(lpos:lpos)
      IF(c.EQ.',')THEN
         code=coma
      ELSEIF(c.EQ.'(')THEN
         code=left
      ELSEIF(c.EQ.')')THEN
         code=right
      ELSEIF(c.EQ.'+')THEN
         code=add
      ELSEIF(c.EQ.'-')THEN
         d=rowb(lpos-1:lpos-1) ! get the character before the minus
         IF(d.EQ.'='.OR.d.EQ.'('.OR.d.EQ.',')THEN ! unary minus
* Convert '-' into '-1*'; send '-1' now and '*' next time
            code=rnum ! real constant
            nrnum=nrnum+1
            rnums(nrnum)=-1d0
            val=nrnum
            pend=.TRUE. ! next time send a '*'
         ELSE ! binary minus
            code=sub
         ENDIF
      ELSEIF(c.EQ.'*')THEN
         IF(rowb(lpos+1:lpos+1).EQ.'*')THEN
            code=pow
            lpos=lpos+1
         ELSE
            code=mul
         ENDIF
      ELSEIF(c.EQ.'/')THEN
         code=div
      ELSEIF(c.EQ.'=')THEN
         code=assign
      ELSEIF(digito(c).OR.c.EQ.'.')THEN ! a number begins
         nume=.TRUE.
         cad=c ! it'll be completed below
      ELSEIF(letra(c))THEN ! variable or intrinsic
         DOWHILE(letra(c).OR.digito(c)) ! complete the name
            cad=cad(1:INDEX(cad,' ')-1)//c
            lpos=lpos+1
            c=rowb(lpos:lpos)
         ENDDO
         lpos=lpos-1
* Identify the symbol
* Intrinsic
         code=intr
         DO i=1,24 ! convert specific names into generic names
            IF(ma(cad).EQ.specific(i))cad=cad(2:INDEX(cad,' ')-1)
         ENDDO
         DO i=1,maxstd
            val=i
            IF(ma(cad).EQ.intrin(i))RETURN ! it is an intrinsic
         ENDDO
* Derivable variable
         code=rvar
         DO i=1,nrvar
            val=i
            IF(ma(cad).EQ.ma(rvars(i)))THEN ! derivable variable
               lpos=lpos+1
               c=rowb(lpos:lpos)
               IF(c.EQ.'(')THEN ! match without (), but send with ()
                  DOWHILE(c.NE.')') ! complete until ')'
                     cad=cad(1:INDEX(cad,' ')-1)//c
                     lpos=lpos+1
                     c=rowb(lpos:lpos)
                  ENDDO
                  cad=cad(1:INDEX(cad,' ')-1)//c ! right parenthesis
                  nrvar=nrvar+1
                  val=nrvar ! add name with parentheses
                  rvars(val)=cad
               ELSE
                  lpos=lpos-1
               ENDIF
               RETURN
            ENDIF
         ENDDO
* Non derivable variable
         code=pvar
         DO i=1,npvar
            val=i
            IF(ma(cad).EQ.ma(pvars(i)))THEN ! non derivable variable
               lpos=lpos+1
               c=rowb(lpos:lpos)
               IF(c.EQ.'(')THEN ! match without (), but send with ()
                  DOWHILE(c.NE.')') ! complete until ')'
                     cad=cad(1:INDEX(cad,' ')-1)//c
                     lpos=lpos+1
                     c=rowb(lpos:lpos)
                  ENDDO
                  cad=cad(1:INDEX(cad,' ')-1)//c ! right parenthesis
                  npvar=npvar+1
                  val=npvar ! add name with parentheses
                  pvars(val)=cad
               ELSE
                  lpos=lpos-1
               ENDIF
               RETURN
            ENDIF
         ENDDO
         PRINT *,cad
         STOP 'variable not loaded (probably undeclared)'
* New line
      ELSEIF(c.EQ.' '.OR.c.EQ.'!')THEN
         code=nline
         lpos=0
      ENDIF
* A number has started
      IF(nume)THEN
         lpos=lpos+1
         c=rowb(lpos:lpos)
         DOWHILE(digito(c))
            cad=cad(1:INDEX(cad,' ')-1)//c
            lpos=lpos+1
            c=rowb(lpos:lpos)
         ENDDO
         IF(c.NE.'.'.AND.ma(c).NE.'D')THEN
* Insert new integer constant
            code=inum
            ninum=ninum+1
            READ(cad,'(i12)')inums(ninum)
            val=ninum
            lpos=lpos-1
            RETURN
         ELSEIF(c.EQ.'.')THEN
            cad=cad(1:INDEX(cad,' ')-1)//c
            lpos=lpos+1
            c=rowb(lpos:lpos)
            DOWHILE(digito(c))
               cad=cad(1:INDEX(cad,' ')-1)//c
               lpos=lpos+1
               c=rowb(lpos:lpos)
            ENDDO
         ENDIF
         IF(ma(c).EQ.'D')THEN
            cad=cad(1:INDEX(cad,' ')-1)//c
            lpos=lpos+1
            c=rowb(lpos:lpos)
            IF(c.EQ.'+'.OR.c.EQ.'-')THEN
               cad=cad(1:INDEX(cad,' ')-1)//c
               lpos=lpos+1
               c=rowb(lpos:lpos)
            ENDIF
            DOWHILE(digito(c))
               cad=cad(1:INDEX(cad,' ')-1)//c
               lpos=lpos+1
               c=rowb(lpos:lpos)
            ENDDO
         ENDIF
* Real number completed; insert it
         code=rnum
         nrnum=nrnum+1
         rnums(nrnum)=ator(cad)
         val=nrnum
         lpos=lpos-1
      ENDIF
      END

************************************************************************
      SUBROUTINE cierra(npi,pila,var)
      INCLUDE 'smart.par'
      INTEGER npi,pila(piladep),var
* Write to file the pieces of the operations and their derivatives
* Generate new variables as needed.
************************************************************************
      INCLUDE 'smart.com'
      CHARACTER s2*40,s3*40,s4*40,s5*40,q6*6,s1*61
      INTEGER x,j2,j3,j4,j5,aux,j1,naux,nsvar,k,kfin
      DATA q6/'(6x,a)'/
      DATA nsvar/0/
      SAVE q6,nsvar

      INTEGER pas,veloc
      COMMON /pasada/pas,veloc

* Generate left member
      IF(var.EQ.svar)THEN ! new variable
         nsvar=nsvar+1
         IF(nsvar.GT.maxnew)STOP 'increase maxnew in smart.par'
         CALL itoa(nsvar,s2,j2)
         IF(pas.EQ.1)THEN
            svars(nsvar)='o('//s2(1:j2)//')' ! new real variable
         ELSE
            svars(nsvar)='oq('//s2(1:j2)//')' ! new real variable
         ENDIF
         aux=svar
         naux=nsvar
      ELSE ! rvar
         aux=rvar
         naux=pila(npi-6)
      ENDIF
* Retrieve name for a variable or value for a constant
      x=pila(npi-2) ! operation
      CALL build(pila(npi-4),pila(npi-3),s3,j3) ! left operand
      IF(x.NE.noop)CALL build(pila(npi-1),pila(npi),s4,j4)   ! right op
      CALL build(aux,naux,s5,j5)                ! left member
      s1=s5(1:j5)//'='//s3(1:j3) ! build the first part of the sentence
      j1=INDEX(s1,' ')-1
* Write the sentence
      IF(x.EQ.noop)THEN
         WRITE(60,q6)s1(1:j1) ! only a left operand
      ELSEIF(x.EQ.add)THEN
         WRITE(60,q6)s1(1:j1)//'+'//s4(1:j4)
      ELSEIF(x.EQ.sub)THEN
         WRITE(60,q6)s1(1:j1)//'-'//s4(1:j4)
      ELSEIF(x.EQ.mul)THEN
         WRITE(60,q6)s1(1:j1)//'*'//s4(1:j4)
      ELSEIF(x.EQ.div)THEN
         WRITE(60,q6)s1(1:j1)//'/'//s4(1:j4)
      ELSEIF(x.EQ.pow)THEN
         WRITE(60,q6)s1(1:j1)//'**'//s4(1:j4)
      ELSEIF(x.EQ.left)THEN ! intrinsic function, 1 argument
         WRITE(60,q6)s1(1:j1)//'('//s4(1:j4)//')'
      ELSEIF(x.EQ.coma)THEN ! ATAN2
         WRITE(60,q6)s5(1:j5)//'=ATAN2('//s3(1:j3)//','//s4(1:j4)//')'
      ELSE
         STOP 'unknown operation'
      ENDIF
* Derivation
      IF(pas.EQ.1.OR.(pas.NE.1.AND.veloc.EQ.0))THEN
* Accelerations or potential indep. of v: derivatives wrt x only
         kfin=ndim
      ELSE
         kfin=2*ndim ! variationals & pot(v): derivatives wrt x & v
      ENDIF
      DO k=1,kfin
* pila(npi-2) ! operation
* pila(npi-4) ! type (code) of left operand
* pila(npi-3) ! # left operand
* pila(npi-1) ! type (code) of right operand
* pila(npi)   ! # right operand
* aux,naux    ! to store the result
         CALL der(k,pila(npi-2),pila(npi-4),pila(npi-3),pila(npi-1),
     &    pila(npi),aux,naux)
      ENDDO
      WRITE(60,*)
* Replace left operand with result
      npi=npi-3
      pila(npi-1)=aux
      pila(npi)=naux
      IF(x.EQ.coma)THEN ! go back three more slots if ATAN2
         npi=npi-3
         pila(npi-1)=aux
         pila(npi)=naux
      ENDIF
      END

************************************************************************
      SUBROUTINE der(ixx,op,cizq,nizq,cder,nder,cres,nres)
      INTEGER ixx,op,cizq,nizq,cder,nder,cres,nres
* Write the derivative of the last operation
* ixx = # of the variable wrt which the derivation is done
* op = code of operation (add, sub, mul, div, pow, left (=intrinsic))
* cizq = code left operand (inum, rnum, rvar, intr)
* nizq = # of left operand (variable/constant/intrinsic)
* cder = code of right operand (inum, rnum, rvar)
* nder = # of right operand (variable/constant/argument)
* cres = rvar/svar
* nres = # of variable to hold the derivative
************************************************************************
      INCLUDE 'smart.par'

      INTEGER j1,j2,j3,j1d,j2d,j3d,jix,ix,j4,jr,ip,car
      CHARACTER s1*40,s2*40,s3*40,q6*6,s1d*40,s2d*40,s3d*40,row*144,
     & six*30,s4*40,ma*1369,deno*90
      LOGICAL uno
      DATA q6/'(6x,a)'/

      INTEGER pas,veloc
      COMMON /pasada/pas,veloc

      ix=ixx ! local variable
* Variables
      CALL build(cizq,nizq,s1,j1) ! Name of the left operand
      IF(op.NE.noop)CALL build(cder,nder,s2,j2) ! right operand
      CALL build(cres,nres,s3,j3) ! Name of the resulting variable
* Left operand
      CALL itoa(ix,six,jix)
      IF(ma(s1(1:2)).EQ.'X(')THEN ! dxi/dxj
         READ(s1(3:INDEX(s1,')')-1),'(i12)')car
         IF(car.EQ.ix)THEN
            s1d='1d0'
         ELSE
            s1d='0d0'
         ENDIF
      ELSEIF(pas.EQ.1.AND.ma(s1).EQ.'POT')THEN ! dpot/dx=acc
         s1d='acc('//six(1:jix)//')'
      ELSEIF(pas.NE.1.AND.ma(s1(1:4)).EQ.'ACC(')THEN ! dacc/dx=var
         s1d='var('//s1(5:INDEX(s1,')')-1)//','//six(1:jix)//')'
      ELSEIF(cizq.EQ.rvar)THEN ! drvar/dx
         ip=INDEX(s1,'(')
         IF(ip.EQ.0)THEN
            s1d='o'//s1(1:j1)//'('//six(1:jix)//')'
         ELSE
            s1d='o'//s1(1:j1-1)//','//six(1:jix)//')'
         ENDIF
      ELSEIF(pas.EQ.1.AND.cizq.EQ.svar)THEN ! do()/dx
         s1d='o'//s1(1:j1-1)//','//six(1:jix)//')'
      ELSEIF(pas.NE.1.AND.cizq.EQ.svar)THEN ! do()/dx
         s1d='oq'//s1(1:j1-1)//','//six(1:jix)//')'
      ELSE ! other
         s1d='0d0'
      ENDIF
      j1d=INDEX(s1d,' ')-1
* Right operand
      IF(op.NE.noop)THEN
         IF(ma(s2(1:2)).EQ.'X(')THEN ! same as before
            READ(s2(3:INDEX(s2,')')-1),'(i12)')car
            IF(car.EQ.ix)THEN
               s2d='1d0'
            ELSE
               s2d='0d0'
            ENDIF
         ELSEIF(pas.EQ.1.AND.ma(s2).EQ.'POT')THEN
            s2d='acc('//six(1:jix)//')'
         ELSEIF(pas.NE.1.AND.ma(s2(1:4)).EQ.'ACC(')THEN 
            s2d='var('//s2(5:INDEX(s2,')')-1)//','//six(1:jix)//')'
         ELSEIF(cder.EQ.rvar)THEN
            ip=INDEX(s2,'(')
            IF(ip.EQ.0)THEN
               s2d='o'//s2(1:j2)//'('//six(1:jix)//')'
            ELSE
               s2d='o'//s2(1:j2-1)//','//six(1:jix)//')'
            ENDIF
         ELSEIF(pas.EQ.1.AND.cder.EQ.svar)THEN
            s2d='o'//s2(1:j2-1)//','//six(1:jix)//')'
         ELSEIF(pas.NE.1.AND.cder.EQ.svar)THEN
            s2d='oq'//s2(1:j2-1)//','//six(1:jix)//')'
         ELSE
            s2d='0d0'
         ENDIF
         j2d=INDEX(s2d,' ')-1
      ENDIF
* Left member
      IF(pas.EQ.1.AND.ma(s3).EQ.'POT')THEN
         s3d='acc('//six(1:jix)//')='
      ELSEIF(pas.NE.1.AND.ma(s3(1:4)).EQ.'ACC(')THEN
         s3d='var('//s3(5:INDEX(s3,')')-1)//','//six(1:jix)//')='
      ELSEIF(cres.EQ.rvar)THEN
         ip=INDEX(s3,'(')
         IF(ip.EQ.0)THEN
            s3d='o'//s3(1:j3)//'('//six(1:jix)//')='
         ELSE
            s3d='o'//s3(1:j3-1)//','//six(1:jix)//')='
         ENDIF
      ELSEIF(pas.EQ.1)THEN
         s3d='o'//s3(1:j3-1)//','//six(1:jix)//')='
      ELSE
         s3d='oq'//s3(1:j3-1)//','//six(1:jix)//')='
      ENDIF
      j3d=INDEX(s3d,' ')-1
* Sentences of the derivatives; simplify when possible
      IF(op.EQ.noop)THEN
         row=s1d(1:j1d) ! only a left operand
      ELSEIF(op.EQ.add)THEN
         IF(s1d.EQ.'0d0'.AND.s2d.EQ.'0d0')THEN
            row='0d0'
         ELSEIF(s1d.EQ.'0d0')THEN
            row=s2d(1:j2d)
         ELSEIF(s2d.EQ.'0d0')THEN
            row=s1d(1:j1d)
         ELSE
            row=s1d(1:j1d)//'+'//s2d(1:j2d)
         ENDIF
      ELSEIF(op.EQ.sub)THEN
         IF(s1d.EQ.'0d0'.AND.s2d.EQ.'0d0')THEN
            row='0d0'
         ELSEIF(s1d.EQ.'0d0')THEN
            row='-'//s2d(1:j2d)
         ELSEIF(s2d.EQ.'0d0')THEN
            row=s1d(1:j1d)
         ELSE
            row=s1d(1:j1d)//'-'//s2d(1:j2d)
         ENDIF
      ELSEIF(op.EQ.mul)THEN
         IF(s1d.EQ.'0d0'.AND.s2d.EQ.'0d0')THEN
            row='0d0'
         ELSEIF(s1d.EQ.'0d0'.AND.s2d.EQ.'1d0')THEN
            row=s1(1:j1)
         ELSEIF(s1d.EQ.'1d0'.AND.s2d.EQ.'0d0')THEN
            row=s2(1:j2)
         ELSEIF(s1d.EQ.'1d0'.AND.s2d.EQ.'1d0')THEN
            row=s2(1:j2)//'+'//s1(1:j1)
         ELSEIF(s1d.EQ.'0d0')THEN
            row=s1(1:j1)//'*'//s2d(1:j2d)
         ELSEIF(s2d.EQ.'0d0')THEN
            row=s1d(1:j1d)//'*'//s2(1:j2)
         ELSEIF(s1d.EQ.'1d0')THEN
            row=s2(1:j2)//'+'//s1(1:j1)//'*'//s2d(1:j2d)
         ELSEIF(s2d.EQ.'1d0')THEN
            row=s1d(1:j1d)//'*'//s2(1:j2)//'+'//s1(1:j1)
         ELSE
            row=s1d(1:j1d)//'*'//s2(1:j2)//'+'//s1(1:j1)//'*'//
     &       s2d(1:j2d)
         ENDIF
      ELSEIF(op.EQ.div)THEN
         IF(s1d.EQ.'0d0'.AND.s2d.EQ.'0d0')THEN
            row='0d0'
         ELSEIF(s1d.EQ.'0d0'.AND.s2d.EQ.'1d0')THEN
            row='-'//s1(1:j1)//'/'//s2(1:j2)//'**2'
         ELSEIF(s1d.EQ.'1d0'.AND.s2d.EQ.'0d0')THEN
            row='1d0/'//s2(1:j2)
         ELSEIF(s1d.EQ.'1d0'.AND.s2d.EQ.'1d0')THEN
            row='('//s2(1:j2)//'-'//s1(1:j1)//')/'//s2(1:j2)//'**2'
         ELSEIF(s1d.EQ.'0d0')THEN
            row='-'//s1(1:j1)//'*'//s2d(1:j2d)//'/'//s2(1:j2)//'**2'
         ELSEIF(s2d.EQ.'0d0')THEN
            row=s1d(1:j1d)//'/'//s2(1:j2)
         ELSEIF(s1d.EQ.'1d0')THEN
            row='('//s2(1:j2)//'-'//s1(1:j1)//'*'//s2d(1:j2d)//')/'//
     &       s2(1:j2)//'**2'
         ELSEIF(s2d.EQ.'1d0')THEN
            row='('//s1d(1:j1d)//'*'//s2(1:j2)//'-'//s1(1:j1)//')/'//
     &       s2(1:j2)//'**2'
         ELSE
            row='('//s1d(1:j1d)//'*'//s2(1:j2)//'-'//
     &       s1(1:j1)//'*'//s2d(1:j2d)//')/'//s2(1:j2)//'**2'
         ENDIF
      ELSEIF(op.EQ.pow)THEN
         CALL menosuno(cder,s2,uno,s4,j4) ! exp - 1; uno=T if exp-1=1
         IF(s1d.EQ.'0d0'.AND.s2d.EQ.'0d0')THEN
            row='0d0'
         ELSEIF(s1d.EQ.'0d0'.AND.s2d.EQ.'1d0')THEN
            row=s1(1:j1)//'**'//s2(1:j2)//'*LOG(ABS('//s1(1:j1)//'))'
         ELSEIF(s1d.EQ.'1d0'.AND.s2d.EQ.'0d0')THEN
            IF(uno)THEN
               row=s2(1:j2)//'*'//s1(1:j1)
            ELSE
               row=s2(1:j2)//'*'//s1(1:j1)//'**'//s4(1:j4)
            ENDIF
         ELSEIF(s1d.EQ.'1d0'.AND.s2d.EQ.'1d0')THEN
            IF(uno)THEN
               row=s2(1:j2)//'*'//s1(1:j1)//'+'//s1(1:j1)//'**'//
     &          s2(1:j2)//'*LOG(ABS('//s1(1:j1)//'))'
            ELSE
               row=s2(1:j2)//'*'//s1(1:j1)//'**'//s4(1:j4)//'+'//
     &          s1(1:j1)//'**'//s2(1:j2)//'*LOG(ABS('//s1(1:j1)//'))'
            ENDIF
         ELSEIF(s1d.EQ.'0d0')THEN
            row=s1(1:j1)//'**'//s2(1:j2)//'*LOG(ABS('//s1(1:j1)//
     &       '))*'//s2d(1:j2d)
         ELSEIF(s2d.EQ.'0d0')THEN
            IF(uno)THEN
               row=s2(1:j2)//'*'//s1(1:j1)//'*'//s1d(1:j1d)
            ELSE
               row=s2(1:j2)//'*'//s1(1:j1)//'**'//s4(1:j4)//'*'//
     &          s1d(1:j1d)
            ENDIF
         ELSEIF(s1d.EQ.'1d0')THEN
            IF(uno)THEN
               row=s2(1:j2)//'*'//s1(1:j1)//'+'//s1(1:j1)//'**'//
     &          s2(1:j2)//'*LOG(ABS('//s1(1:j1)//'))*'//s2d(1:j2d)
            ELSE
               row=s2(1:j2)//'*'//s1(1:j1)//'**'//s4(1:j4)//'+'//
     &          s1(1:j1)//'**'//s2(1:j2)//'*LOG(ABS('//s1(1:j1)//
     &          '))*'//s2d(1:j2d)
            ENDIF
         ELSEIF(s2d.EQ.'1d0')THEN
            IF(uno)THEN
               row=s2(1:j2)//'*'//s1(1:j1)//'*'//s1d(1:j1d)//'+'//
     &          s1(1:j1)//'**'//s2(1:j2)//'*LOG(ABS('//s1(1:j1)//'))'
            ELSE
               row=s2(1:j2)//'*'//s1(1:j1)//'**'//s4(1:j4)//'*'//
     &          s1d(1:j1d)//'+'//s1(1:j1)//'**'//s2(1:j2)//'*LOG(ABS('//
     &          s1(1:j1)//'))'
            ENDIF
         ELSE
            IF(uno)THEN
               row=s2(1:j2)//'*'//s1(1:j1)//'*'//s1d(1:j1d)//'+'//
     &          s1(1:j1)//'**'//s2(1:j2)//'*LOG(ABS('//s1(1:j1)//
     &          '))*'//s2d(1:j2d)
            ELSE
               row=s2(1:j2)//'*'//s1(1:j1)//'**'//s4(1:j4)//'*'//
     &          s1d(1:j1d)//'+'//s1(1:j1)//'**'//s2(1:j2)//'*LOG(ABS('//
     &          s1(1:j1)//'))*'//s2d(1:j2d)
            ENDIF
         ENDIF
      ELSEIF(op.EQ.left)THEN ! left ( = intrinsic except ATAN2
         IF(s2d(1:j2d).EQ.'0d0')THEN
            row='0d0'
         ELSEIF(nizq.EQ.1)THEN ! ABS
            row='ABS('//s2(1:j2)//')/'//s2(1:j2)//'*'//s2d(1:j2d)
         ELSEIF(nizq.EQ.2)THEN ! SQRT
            row='0.5d0/SQRT('//s2(1:j2)//')*'//s2d(1:j2d)
         ELSEIF(nizq.EQ.3)THEN ! EXP
            row='EXP('//s2(1:j2)//')*'//s2d(1:j2d)
         ELSEIF(nizq.EQ.4)THEN ! LOG
            row=s2d(1:j2d)//'/'//s2(1:j2)
         ELSEIF(nizq.EQ.5)THEN ! LOG10
            row=s2d(1:j2d)//'/('//s2(1:j2)//'*LOG(10d0))'
         ELSEIF(nizq.EQ.6)THEN ! SIN
            row='COS('//s2(1:j2)//')*'//s2d(1:j2d)
         ELSEIF(nizq.EQ.7)THEN ! COS
            row='-SIN('//s2(1:j2)//')*'//s2d(1:j2d)
         ELSEIF(nizq.EQ.8)THEN ! TAN
            row=s2d(1:j2d)//'/COS('//s2(1:j2)//')**2'
         ELSEIF(nizq.EQ.9)THEN ! ASIN
            row=s2d(1:j2d)//'/SQRT(1d0-'//s2(1:j2)//'**2)'
         ELSEIF(nizq.EQ.10)THEN ! ACOS
            row='-'//s2d(1:j2d)//'/SQRT(1d0-'//s2(1:j2)//'**2)'
         ELSEIF(nizq.EQ.11)THEN ! ATAN
            row=s2d(1:j2d)//'/(1d0+'//s2(1:j2)//'**2)'
         ELSEIF(nizq.EQ.12)THEN ! SINH
            row='COSH('//s2(1:j2)//')*'//s2d(1:j2d)
         ELSEIF(nizq.EQ.13)THEN ! COSH
            row='SINH('//s2(1:j2)//')*'//s2d(1:j2d)
         ELSEIF(nizq.EQ.14)THEN ! TANH
            row=s2d(1:j2d)//'/COSH('//s2(1:j2)//')**2'
         ENDIF
      ELSEIF(op.EQ.coma)THEN ! ATAN2
         deno='/('//s1(1:j1)//'**2+'//s2(1:j2)//'**2)'
         IF(s1d.EQ.'0d0'.AND.s2d.EQ.'0d0')THEN
            row='0d0'
         ELSEIF(s1d.EQ.'0d0'.AND.s2d.EQ.'1d0')THEN
            row='-'//s1(1:j1)//deno
         ELSEIF(s1d.EQ.'1d0'.AND.s2d.EQ.'0d0')THEN
            row=s2(1:j2)//deno
         ELSEIF(s1d.EQ.'1d0'.AND.s2d.EQ.'1d0')THEN
            row='('//s2(1:j2)//'-'//s1(1:j1)//')'//deno
         ELSEIF(s1d.EQ.'0d0')THEN
            row='-'//s1(1:j1)//'*'//s2d(1:j2d)//deno
         ELSEIF(s2d.EQ.'0d0')THEN
            row=s2(1:j2)//'*'//s1d(1:j1d)//deno
         ELSEIF(s1d.EQ.'1d0')THEN
            row='('//s2(1:j2)//'-'//s1(1:j1)//'*'//s2d(1:j2d)//')'//deno
         ELSEIF(s2d.EQ.'1d0')THEN
            row='('//s2(1:j2)//'*'//s1d(1:j1d)//'-'//s1(1:j1)//')'//deno
         ELSE
            row='('//s2(1:j2)//'*'//s1d(1:j1d)//'-'//
     &       s1(1:j1)//'*'//s2d(1:j2d)//')'//deno
         ENDIF
      ENDIF
* Write sentence
      jr=INDEX(row,' ')-1
      IF(j3d+jr.LE.66)THEN
         WRITE(60,q6)s3d(1:j3d)//row(1:INDEX(row,' ')-1)
      ELSE ! long lines
         WRITE(60,q6)s3d(1:j3d)//row(1:66-j3d)
         WRITE(60,*)'    &'//row(67-j3d:jr)
      ENDIF
      END

************************************************************************
      SUBROUTINE menosuno(n,s0,uno,s,ju)
      INTEGER n,ju
      LOGICAL uno
      CHARACTER s0*(*),s*(*)
* Builds a string s corresponding to s0-1
* uno = .TRUE. if s = 1
************************************************************************
      INCLUDE 'smart.par'
      INTEGER iexp
      DOUBLE PRECISION rexp,ator

      uno=.TRUE.
      IF(n.EQ.inum)THEN ! numeric integer exponent
         READ(s0,'(i12)')iexp
         iexp=iexp-1
         IF(iexp.EQ.1)RETURN
         CALL itoa(iexp,s,ju) ! write s = exp-1
      ELSEIF(n.EQ.pvar)THEN ! non-numeric integer exponent
         s='('//s0(1:INDEX(s0,' ')-1)//'-1)'
      ELSEIF(n.EQ.rnum)THEN ! numeric real exponent
         rexp=ator(s0)-1d0
         IF(rexp.EQ.1d0)RETURN
         WRITE(s,'(d22.16)')rexp
      ELSE ! non-numeric real exponent
         s='('//s0(1:INDEX(s0,' ')-1)//'-1d0)'
      ENDIF
      ju=INDEX(s,' ')-1
      uno=.FALSE.
      END

************************************************************************
      FUNCTION ator(a)
      DOUBLE PRECISION ator
      CHARACTER a*(*)
* Reads the real number stored in a.
* Standard F77 forbids the use of list directed I/O on internal files.
************************************************************************
      INTEGER lon,dot,dec
      CHARACTER aux*10,slon*2,sdec*2
      lon=INDEX(a,' ')-1
      dot=INDEX(a,'.')
      IF(dot.EQ.0)THEN
         dec=0
      ELSE
         dec=INDEX(a,'D')
         IF(dec.EQ.0)THEN
            dec=INDEX(a,'d')
            IF(dec.EQ.0)THEN
               dec=INDEX(a,'+')
               IF(dec.EQ.0)THEN
                  dec=INDEX(a,'-')
               ENDIF
            ENDIF
         ENDIF
         IF(dec.NE.0)dec=dec-dot-1
      ENDIF
      WRITE(slon,'(i2)')lon
      WRITE(sdec,'(i2)')dec
      aux='(d'//slon//'.'//sdec//')'
      READ(a,aux)ator
      END

************************************************************************
      SUBROUTINE itoa(n,s,ju)
      INTEGER n,ju
      CHARACTER s*(*)
* Writes the integer 'n' in the first 'ju' characters of 's'
************************************************************************
      INTEGER i
      WRITE(s,'(i12)')n
      DO i=1,12 
         IF(s(i:i).NE.' ')THEN
            s=s(i:12)
            ju=12-i+1
            RETURN
         ENDIF
      ENDDO
      END

************************************************************************
      SUBROUTINE build(code,i,s,ju)
      INTEGER i,ju,code
      CHARACTER*(*) s
* Recovers alphanumeric name of constant/variable/intrinsic type 'code'
* with order number 'i'
************************************************************************
      INCLUDE 'smart.par'
      INCLUDE 'smart.com'

      CHARACTER intrin(maxstd)*5
      DATA intrin/'ABS','SQRT','EXP','LOG','LOG10','SIN','COS','TAN',
     & 'ASIN','ACOS','ATAN','SINH','COSH','TANH','ATAN2'/

      IF(code.EQ.inum)THEN ! integer constant
         CALL itoa(inums(i),s,ju)
      ELSEIF(code.EQ.rnum)THEN ! real constant
         WRITE(s,'(d22.16)')rnums(i)
      ELSEIF(code.EQ.rvar)THEN ! real variable
         s=rvars(i)
      ELSEIF(code.EQ.svar)THEN ! new real variable
         s=svars(i)
      ELSEIF(code.EQ.pvar)THEN ! parameter
         s=pvars(i)
      ELSEIF(code.EQ.intr)THEN ! intrinsic function
         s=intrin(i)
      ELSE
         PRINT *,code
         STOP 'unknown code'
      ENDIF
      ju=INDEX(s,' ')-1
      END

************************************************************************
      SUBROUTINE encabezado()
* Write header of the acceleration/variational routine
* Store all the variables as pvar or rvar.
************************************************************************
      INCLUDE 'smart.par'
      INCLUDE 'smart.com'

      INTEGER nrvar,npvar
      COMMON /fijos/nrvar,npvar ! shared with lexi

      INTEGER pas,veloc
      COMMON /pasada/pas,veloc

      CHARACTER s*40,s2*40,row*1369,q6*6,c*1,cad*40,ma*1369,mxn*12
      CHARACTER aun(maxsym)*40,vdim(maxsym)*40,qdim(maxsym)*40,saux*12,
     & ivdim*40,iqdim*40,ip*40,su*40
      LOGICAL pili,letra,digito,decla,ban,prim
      INTEGER i,j,j2,k,ig,naun,l,lant,id1,id2,id,idim,iq,eti,cl,ju,vd,
     & icar1,icar2,ix,ir,iu,ij
      DATA q6/'(6x,a)'/

      CHARACTER intrin(maxstd)*5
      DATA intrin/'ABS','SQRT','EXP','LOG','LOG10','SIN','COS','TAN',
     & 'ASIN','ACOS','ATAN','SINH','COSH','TANH','ATAN2'/
      CHARACTER notallow(37)*6
      DATA notallow/'INT','IFIX','IDINT','REAL','FLOAT','SNGL','DBLE',
     & 'CMPLX','DPROD','AIMAG','CONJG','AINT','DINT','ANINT','DNINT',
     & 'NINT','IDNINT','SIGN','ISIGN','DSIGN','DIM','IDIM','DDIM','MOD',
     & 'AMOD','DMOD','MAX','MAX0','AMAX1','DMAX1','AMAX0','MAX1','MIN0',
     & 'AMIN1','DMIN1','AMIN0','MIN1'/
      CHARACTER specific(24)*6
      DATA specific/'IABS','DABS','CABS','DSQRT','CSQRT','DEXP','CEXP',
     & 'ALOG','DLOG','CLOG','ALOG10','DLOG10','DSIN','CSIN','DCOS',
     & 'CCOS','DTAN','DASIN','DACOS','DATAN','DSINH','DCOSH','DTANH',
     & 'DATAN2'/

      letra(c)=('A'.LE.c.AND.c.LE.'Z').OR.('a'.LE.c.AND.c.LE.'z')
      digito(c)='0'.LE.c.AND.c.LE.'9'

* Search for dimensions among the declarations of DOUBLE PRECISION
* variables, for later use.
      idim=1
      vdim(1)=' '
      CALL readl(0,row,eti,cl) ! 0 = read only, don't write
      DOWHILE(ma(row).NE.'END')
* Verify DIMENSION and IMPLICIT are not used
         id1=INDEX(ma(row),'DIMENSION')
         IF(id1.EQ.1)STOP'The DIMENSION declaration should not be used.'
         id1=INDEX(ma(row),'IMPLICIT')
         IF(id1.EQ.1)STOP'The IMPLICIT declaration should not be used.'
* Verify REAL*4 variables are not used
         id1=INDEX(ma(row),'REAL')
         IF(id1.EQ.1.AND.(row(5:5).NE.'*'.OR.row(5:6).EQ.'*4'))
     &    STOP'REAL*4 variables should not be used.'
* Begin searching
         id1=INDEX(ma(row),'DOUBLEPRECISION')
         id2=INDEX(ma(row),'REAL*8')
         IF(id1.EQ.1.OR.id2.EQ.1)THEN
            IF(id1.EQ.1)THEN
               id=16
            ELSE
               id=7
            ENDIF
            DOWHILE(id.LE.1369)
               c=row(id:id)
               IF(letra(c).OR.digito(c))THEN
* Verify extension of the variable's name
                  IF(INDEX(vdim(idim),' ').EQ.0)STOP 'Names of'//
     &             ' variables should be less than 40 characters long.'
                  vdim(idim)=vdim(idim)(1:INDEX(vdim(idim),' ')-1)//c
               ELSEIF(c.EQ.'(')THEN
                  iq=INDEX(row(id:1369),')')
                  qdim(idim)=row(id:id+iq-1)
* Verify the extension of name+dimensions
                  icar1=INDEX(vdim(idim),' ')-1
                  icar2=INDEX(qdim(idim),' ')-1
                  IF(icar1.LT.0.OR.icar2.LT.0.OR.icar1+icar2.GT.40)STOP
     &             'Names of variables, including their dimensions, '//
     &             'should be less than 40 characters long.'
                  idim=idim+1
                  IF(idim.GT.maxsym)STOP'increase maxsym in smart.par'
                  vdim(idim)=' '
                  id=id+iq
               ELSE
                  vdim(idim)=' '
               ENDIF
               id=id+1
            ENDDO
         ENDIF
* Verify extension of INTEGER variables' names
         ivdim=' '
         id1=INDEX(ma(row),'INTEGER')
         IF(id1.EQ.1)THEN
            id=8
            DOWHILE(id.LE.1369)
               c=row(id:id)
               IF(letra(c).OR.digito(c))THEN
                  IF(INDEX(ivdim,' ').EQ.0)STOP 'Names of'//
     &             ' variables should be less than 40 characters long.'
                  ivdim=ivdim(1:INDEX(ivdim,' ')-1)//c
               ELSEIF(c.EQ.'(')THEN
                  iq=INDEX(row(id:1369),')')
                  iqdim=row(id:id+iq-1)
                  icar1=INDEX(ivdim,' ')-1
                  icar2=INDEX(iqdim,' ')-1
                  IF(icar1.LT.0.OR.icar2.LT.0.OR.icar1+icar2.GT.40)STOP
     &             'Names of variables, including their dimensions, '//
     &             'should be less than 40 characters long.'
                  ivdim=' '
                  id=id+iq
               ELSE
                  ivdim=' '
               ENDIF
               id=id+1
            ENDDO
         ENDIF
         CALL readl(0,row,eti,cl)
      ENDDO
      idim=idim-1
      REWIND(50)

* Function and independent variables plus t
      IF(pas.EQ.1)THEN
         rvars(1)='pot'
      ELSE
         rvars(1)='acc'
      ENDIF
      rvars(2)='x'
      pvars(1)='t'
      nrvar=2
      npvar=1
* Search for derivable and non-derivable variables
      CALL readl(0,row,eti,cl)
* Search until the END sentence
      prim=.TRUE.
      DOWHILE(ma(row).NE.'END')
* Look for the '='
         ig=INDEX(row,'=')
* Shouldn't be a comment line nor an OPEN etc. with '='
         IF(ig.NE.0.AND..NOT.pili(row))THEN
* Look for variables in the line
            naun=0
            id=1
            DOWHILE(id.LE.1369)
               c=row(id:id)
               IF(letra(c))THEN ! found a possible variable
                  lant=id-1 ! store the position of previous token
                  cad=' '
                  DOWHILE(letra(c).OR.digito(c)) ! complete the name
                     cad=cad(1:INDEX(cad,' ')-1)//c
                     id=id+1
                     c=row(id:id)
                  ENDDO
* Verify the presence of not allowed intrinsics
                  DO i=1,37
                     IF(ma(cad).EQ.notallow(i))THEN
                        PRINT *,notallow(i),' intrinsic not allowed.'
                        STOP
                     ENDIF
                  ENDDO
* Discard intrinsics
                  DO i=1,maxstd
                     IF(ma(cad).EQ.intrin(i))GOTO 3
                  ENDDO
                  DO i=1,24
                     IF(ma(cad).EQ.specific(i))GOTO 3
                  ENDDO
* Discard 'e' or 'd' of exponents
                  IF(ma(cad(1:1)).EQ.'E'.OR.ma(cad(1:1)).EQ.'D')THEN
                     IF(lant.NE.0)THEN
                        c=row(lant:lant)
                        IF(digito(c).OR.c.EQ.'.')GOTO 3
                     ENDIF
                  ENDIF
* Verify the presence of statement function statements.
                  IF(pas.EQ.1.AND.naun.EQ.0.AND.row(id:id).EQ.'('.AND.
     &             prim.AND.ma(row(id+1:id+2)).NE.'N,'.AND.
     &             ma(row(id+1:id+2)).NE.'N)'.AND.
     &             .NOT.digito(row(id+1:id+1)))STOP
     &             'Statement function statements are not allowed.'
                  IF(prim)prim=.FALSE.
* Verify the variable does not begin with 'o'
                  IF(pas.EQ.1.AND.ma(cad(1:1)).EQ.'O')STOP
     &             'Variables should not begin with ''o''.'
* Verify that 'x' has a constant index
                  IF(pas.EQ.1.AND.ma(cad).EQ.'X'.AND.
     &             ma(row(id:id)).EQ.'('.AND.
     &             .NOT.digito(row(id+1:id+1)))STOP 'Variable ''x'''//
     &             ' should appear with a numerical index.'
* Store new variables
                  naun=naun+1
                  IF(naun.GT.maxsym)STOP'increase maxsym in smart.par'
                  aun(naun)=cad ! aun(1)=left member

* If the potential depends on velocities, set veloc=1 to derive wrt v
                  IF(ma(aun(naun)).EQ.'X'.AND.veloc.EQ.0)THEN
                     saux=row(id+1:id+INDEX(row(id+1:1369),')')-1)
                     READ(saux,'(i12)',ERR=7)vd
                     IF(vd.GT.ndim)veloc=1
                  ENDIF

3                 CONTINUE
               ENDIF
               id=id+1
            ENDDO
* Verify whether there are functions with derivable variables
            IF(pas.EQ.1)THEN
               DO j=2,naun
                  su=ma(aun(j))
                  ij=INDEX(su,' ')-1
                  iu=INDEX(ma(row),su(1:ij))+ij
                  IF(row(iu:iu).EQ.'(')THEN
                     ip=ma(row(iu:INDEX(row,')')))
                     DO i=1,nrvar
                        su=ma(rvars(i))
                        ij=INDEX(su,' ')-1
                        ix=INDEX(ip,su(1:ij))
                        ir=INDEX(rvars(i),' ')-1
                        IF(ix.NE.0.AND..NOT.letra(ip(ix-1:ix-1))
     &                   .AND..NOT.digito(ip(ix-1:ix-1))
     &                   .AND..NOT.letra(ip(ix+ir:ix+ir))
     &                   .AND..NOT.digito(ip(ix+ir:ix+ir))) 
     &                   STOP 'Auxiliary routines should not contain '//
     &                   'derivable variables.'
                     ENDDO
                  ENDIF
               ENDDO
            ENDIF

* rvar, left: nothing to do
            DO i=1,nrvar
               IF(ma(aun(1)).EQ.ma(rvars(i)))GOTO 1
            ENDDO
* pvar, left: changes to rvar if there is an rvar in the right member
            DO i=1,npvar
               IF(ma(aun(1)).EQ.ma(pvars(i)))THEN
                  DO j=2,naun
                     DO k=1,nrvar
                        IF(ma(aun(j)).EQ.ma(rvars(k)))THEN
                           nrvar=nrvar+1 ! add an rvar
                           rvars(nrvar)=pvars(i) ! =aun(1)
                           DO l=i+1,npvar ! delete the pvar
                              pvars(l-1)=pvars(l)
                           ENDDO
                           npvar=npvar-1
                           GOTO 1
                        ENDIF
                     ENDDO
                  ENDDO
                  GOTO 1 ! an existent pvar: nothing to do
               ENDIF
            ENDDO
* new, left: rvar if it is an 'o' variable
            IF(ma(aun(1)(1:1)).EQ.'O')THEN
               nrvar=nrvar+1 ! add an rvar
               rvars(nrvar)=aun(1)
               GOTO 1
            ENDIF
* new, left: rvar if there's an rvar in the right member, pvar otherwise
            DO j=2,naun
               DO k=1,nrvar
                  IF(ma(aun(j)).EQ.ma(rvars(k)))THEN
                     nrvar=nrvar+1 ! add an rvar
                     rvars(nrvar)=aun(1)
                     GOTO 1
                  ENDIF
               ENDDO
            ENDDO
            npvar=npvar+1
            pvars(npvar)=aun(1)
* new,right: a new pvar
1           DO j=2,naun
               ban=.TRUE.
               DO k=1,nrvar
                  IF(ma(aun(j)).EQ.ma(rvars(k)))ban=.FALSE.
               ENDDO
               DO k=1,npvar
                  IF(ma(aun(j)).EQ.ma(pvars(k)))ban=.FALSE.
               ENDDO
               IF(ban)THEN
                  npvar=npvar+1
                  pvars(npvar)=aun(j)
               ENDIF
            ENDDO
         ENDIF
         CALL readl(0,row,eti,cl)
      ENDDO

* Write header
      REWIND(50)
      CALL readl(1,row,eti,cl)
* Subprogram line
      CALL itoa(2*ndim,s,j) ! 2* number of independent variables
      IF(pas.EQ.1)THEN
         IF(ma(row).NE.'FUNCTIONPOT(T,X,N)')STOP
     &    'The first non-comment line should be FUNCTION pot(t,x,n)'
         WRITE(60,q6)'SUBROUTINE acelera(t,x,n,acc)'
         WRITE(60,q6)'DOUBLE PRECISION acc(n/2)'
      ELSE
         WRITE(60,q6)'SUBROUTINE variac(t,x,dx,n,dax)'
         WRITE(60,q6)'DOUBLE PRECISION dx(n),dax(n/2),var(n/2,n)'
         READ(50,'(a)') ! replace the acc line (is not a dummy here)
         WRITE(60,q6)'DOUBLE PRECISION acc('//s(1:j)//'/2)'
      ENDIF
      CALL readl(1,row,eti,cl)
      DOWHILE(decla(row)) ! original declarations
         CALL writel(cl)
         CALL readl(1,row,eti,cl)
      ENDDO
* New declarations
      WRITE(60,*)
      CALL itoa(maxnew,mxn,ju)
      IF(pas.EQ.1)THEN
         WRITE(60,q6)'INTEGER oi'
         WRITE(60,q6)'DOUBLE PRECISION oo('//mxn(1:ju)//','//
     &    s(1:j)//'),o('//mxn(1:ju)//')'
      ELSE
         WRITE(60,q6)'INTEGER oj'
         WRITE(60,q6)'DOUBLE PRECISION oqoq('//mxn(1:ju)//','//
     &    s(1:j)//'),oq('//mxn(1:ju)//')'
      ENDIF
      DO i=3,nrvar
         DO k=1,idim ! look for dimensioned variables
            IF(ma(rvars(i)).EQ.ma(vdim(k)))THEN
               s2=rvars(i)(1:INDEX(rvars(i),' ')-1)//qdim(k)
               j2=INDEX(s2,' ')-1
* Discard dimensioned variables of the first pass and 'o's'
               IF(pas.NE.1.AND.(ma(s2(1:1)).NE.'O'.OR.
     &          ma(s2(1:2)).EQ.'O('))GOTO 4
               WRITE(60,q6)'DOUBLE PRECISION o'//s2(1:j2-1)//','//
     &          s(1:j)//')'
               GOTO 4
            ENDIF
         ENDDO
         s2=rvars(i)
         j2=INDEX(s2,' ')-1
         IF(pas.NE.1.AND.ma(s2(1:1)).NE.'O')GOTO 4
         WRITE(60,q6)'DOUBLE PRECISION o'//s2(1:j2)//'('//s(1:j)//')'
4        CONTINUE
      ENDDO
* Rewind the last logical line already read
      DO i=1,cl
         BACKSPACE(50)
      ENDDO
      RETURN
7     STOP 'Variable ''x'' should appear with a numerical index.'
      END

************************************************************************
      SUBROUTINE copiar(chau)
      LOGICAL chau
* If the next row in *fun has a '=', it is to be analized; otherwise,
* it is to be copied as it is.
* Writes the final part of the routine.
************************************************************************
      INCLUDE 'smart.par'
      INCLUDE 'smart.com'

      INTEGER nrvar,npvar
      COMMON /fijos/nrvar,npvar

      CHARACTER row*1369,q6*6,ma*1369,rowb*1369,su*40,c*1
      LOGICAL haylog,pili,a1,a2,letra,digito
      INTEGER eti,cl,i,j,ip,iu,ij,ix,ir
      DATA q6/'(6x,a)'/

      INTEGER pas,veloc
      COMMON /pasada/pas,veloc

      letra(c)=('A'.LE.c.AND.c.LE.'Z').OR.('a'.LE.c.AND.c.LE.'z')
      digito(c)='0'.LE.c.AND.c.LE.'9'

      DOWHILE(.TRUE.)
         CALL readl(1,row,eti,cl)
         IF(ma(row).EQ.'END')THEN
* Write final lines
            IF(pas.EQ.1)THEN
               WRITE(60,q6)'DO oi=1,n/2'
               WRITE(60,q6)'   acc(oi)=-acc(oi)'
               WRITE(60,q6)'ENDDO'
               WRITE(60,q6)'END'
            ELSE
               WRITE(60,q6)'DO oi=1,n/2'
               WRITE(60,q6)'   dax(oi)=0d0'
               IF(veloc.EQ.0)THEN
                  WRITE(60,q6)'   DO oj=1,n/2'
               ELSE
                  WRITE(60,q6)'   DO oj=1,n'
               ENDIF
               WRITE(60,q6)'      dax(oi)=dax(oi)+var(oi,oj)*dx(oj)'
               WRITE(60,q6)'   ENDDO'
               WRITE(60,q6)'ENDDO'
               WRITE(60,q6)'END'
            ENDIF
* Copy any possible additional comment line and additional routines
            DOWHILE(.TRUE.)
               READ(50,'(a72)',END=1)row
               WRITE(60,'(a72)')row
            ENDDO
1           CLOSE(50)
            CLOSE(60)
            chau=.TRUE.
            RETURN ! end of the program
         ENDIF
* Verify that IF is paired with THEN
         j=INDEX(row,' ')-1
         IF(ma(row(1:3)).EQ.'IF('.AND.ma(row(j-3:j)).NE.'THEN')STOP
     &    'IF should be paired with THEN.'
* Verify whether there are subroutines with derivable variables
         IF(ma(row(1:4)).EQ.'CALL'.AND.pas.EQ.1)THEN
            ip=INDEX(row,'(')
            iu=INDEX(row,')')
            rowb=ma(row(ip:iu))
            DO i=1,nrvar
               su=ma(rvars(i))
               ij=INDEX(su,' ')-1
               ix=INDEX(rowb,su(1:ij))
               ir=INDEX(rvars(i),' ')-1
               IF(ix.NE.0.AND..NOT.letra(rowb(ix-1:ix-1))
     &          .AND..NOT.digito(rowb(ix-1:ix-1))
     &          .AND..NOT.letra(rowb(ix+ir:ix+ir))
     &          .AND..NOT.digito(rowb(ix+ir:ix+ir))) 
     &          STOP 'Auxiliary routines should not contain '//
     &          'derivable variables.'
            ENDDO
         ENDIF

* Pick lines to analise; copy them otherwise
         a1=pas.EQ.1.AND.INDEX(row,'=').NE.0.AND..NOT.haylog(row).AND.
     &    .NOT.pili(row)
         a2=pas.NE.1.AND.ma(row(1:2)).NE.'O('
     &    .AND.ma(row(1:5)).NE.'OPEN('.AND.
     &    (ma(row(1:1)).EQ.'O'.OR.ma(row(1:4)).EQ.'ACC(')
         IF(a1.OR.a2)THEN
            DO i=1,cl
               BACKSPACE(50)
            ENDDO
            RETURN
         ELSE
            CALL writel(cl)
         ENDIF
      ENDDO
      END

************************************************************************
      FUNCTION haylog(row)
      LOGICAL haylog
      CHARACTER row*1369
* .TRUE. if the line is a pvar assignment
************************************************************************
      INCLUDE 'smart.par'
      INCLUDE 'smart.com'

      INTEGER nrvar,npvar
      COMMON /fijos/nrvar,npvar ! shared with lexi

      INTEGER ig,i
      CHARACTER cad*40,ma*1369

      haylog=.FALSE.
* Find the variable before the '='
      ig=INDEX(row,'=')-1
      cad=' '
      DO i=1,ig
         IF(row(i:i).EQ.'(')GOTO 1 ! don't consider the parentheses
         IF(row(i:i).NE.' ')cad=cad(1:INDEX(cad,' ')-1)//row(i:i)
      ENDDO
* If the variable is a pvar
1     DO i=1,npvar
         IF(ma(pvars(i)).EQ.ma(cad))THEN
            haylog=.TRUE.
            RETURN
         ENDIF
      ENDDO
      END

************************************************************************
      FUNCTION decla(row)
      LOGICAL decla
      CHARACTER row*1369
* .TRUE. if row is a declaration.
************************************************************************
      INTEGER i
      CHARACTER dec(17)*16,ma*1369
      DATA dec/'LOGICAL','INTEGER','REAL','DOUBLEPRECISION',
     & 'COMPLEX','DOUBLECOMPLEX','CHARACTER','COMMON','DIMENSION',
     & 'PARAMETER','DATA','SAVE','EQUIVALENCE','IMPLICIT','EXTERNAL',
     & 'INTRINSIC','FORMAT'/
      decla=.TRUE.
      DO i=1,16
         IF(INDEX(ma(row),dec(i)(1:INDEX(dec(i),' ')-1)).EQ.1)RETURN
      ENDDO
      decla=.FALSE.
      END

************************************************************************
      FUNCTION pili(row)
      LOGICAL pili
      CHARACTER row*1369
* .TRUE. if 'row' is a sentence that can contain a '=' but is
* not an assignment
************************************************************************
      INTEGER nord
      PARAMETER (nord=12)
      INTEGER icad,j,ip,ic,ii,iq
      CHARACTER orden(nord)*10,ma*1369,rowb*1369
      DATA orden/'OPEN','CLOSE','INQUIRE','READ','WRITE','PRINT',
     & 'BACKSPACE','ENDFILE','REWIND','PARAMETER','DATA','DO'/

      pili=.TRUE.
      rowb=ma(row)
      DO j=1,nord
         iq=INDEX(orden(j),' ')-1
         icad=INDEX(rowb,orden(j)(1:iq))
         IF(icad.EQ.1)THEN
            IF(j.LE.nord-2)THEN
               IF(rowb(iq+1:iq+1).EQ.'(')RETURN
            ELSEIF(j.EQ.nord-1)THEN ! DATA w/implied DO
               ii=INDEX(rowb,'=')
               ic=INDEX(rowb,'(')
               IF(ii.NE.0.AND.ic.NE.0.AND.ic.LT.ii)RETURN
            ELSE ! DO
               ii=INDEX(rowb,'=')
               ic=INDEX(rowb,',')
               ip=INDEX(rowb,'(')
               IF(ii.NE.0.AND.ic.NE.0.AND.ip.EQ.0)RETURN
            ENDIF
         ENDIF
      ENDDO
      pili=.FALSE.
      END

************************************************************************
      FUNCTION bl(ecad)
      CHARACTER bl*1369,ecad*(*)
* Eliminates blanks in ecad.
************************************************************************
      INTEGER ib,i
      CHARACTER c*1,cad*1369
* Don't disturb the original string
      cad=ecad
* Last non-blank character
      DO i=LEN(cad),1,-1
         IF(cad(i:i).NE.' ')GOTO 1
      ENDDO
1     ib=i
      DO i=ib,1,-1
         c=cad(i:i)
         IF(c.EQ.' ')THEN ! blanks
            cad(i:ib-1)=cad(i+1:ib)
            cad(ib:ib)=' '
            ib=ib-1
         ENDIF
      ENDDO
      bl=cad
      END
************************************************************************
      FUNCTION ma(ecad)
      CHARACTER ma*1369,ecad*(*)
* Turns ecad to capitals.
************************************************************************
      INTEGER i
      LOGICAL letram
      CHARACTER c*1,cad*1369

      letram(c)='a'.LE.c.AND.c.LE.'z'
* Don't disturb the original string
      cad=ecad
      DO i=1,LEN(ecad)
         c=cad(i:i)
         IF(letram(c))cad(i:i)=CHAR(ICHAR(c)-32) ! to capitals
      ENDDO
      ma=cad
      END

************************************************************************
      SUBROUTINE readl(w,rowb,label,cl)
      CHARACTER rowb*1369
      INTEGER label,cl,w
* Read the next non-comment logical line (i.e. including continuation
* lines).
* Previous comment lines are written if w <> 0.
* Return also a possible label and number of physical lines (cl)
* All blank spaces are deleted.
************************************************************************
      CHARACTER row*72,bl*1369,c*1
      LOGICAL comen,conti

      comen(c)=c.EQ.'C'.OR.c.EQ.'c'.OR.c.EQ.'*'.OR.c.EQ.'!'
      conti(c)=c.NE.' '.AND.c.NE.'0'

      label=0
* Comment or blank lines: write as they are
      READ(50,'(a72)')row
      DOWHILE(comen(row(1:1)).OR.row.EQ.' ')
         IF(w.NE.0)WRITE(60,'(a72)')row
         READ(50,'(a72)')row
      ENDDO
* Label zone
      IF(row(1:5).NE.' ')READ(row(1:5),'(i5)')label
      rowb=bl(row(7:72))
* Complete with possible continuation lines
      READ(50,'(a72)',END=1)row
      cl=1 ! one less than already read -the last one will be backspaced
      DOWHILE(conti(row(6:6)))
         IF(.NOT.comen(row(1:1)).AND.row.NE.' ')
     &    rowb=rowb(1:INDEX(rowb,' ')-1)//bl(row(7:72))
         READ(50,'(a72)',END=1)row
         cl=cl+1
      ENDDO
1     BACKSPACE(50)
      rowb=rowb(1:1368)//' ' ! blank: end of the logical line
      END

************************************************************************
      SUBROUTINE writel(cl)
      INTEGER cl
* Rereads and writes a logical line.
************************************************************************
      INTEGER i
      CHARACTER row*72
      DO i=1,cl
         BACKSPACE(50)
      ENDDO
      DO i=1,cl
         READ(50,'(a72)')row
         WRITE(60,'(a72)')row
      ENDDO
      END

c***********************************************************************
c     SUBROUTINE pilaback(pila,npi)
c     INCLUDE 'smart.par'
c     INTEGER pila(piladep),npi
c* For debugging purposes.
c***********************************************************************
c     INTEGER i,j,p
c     CHARACTER s*40
c     i=1
c     DOWHILE(i.LE.npi)
c        p=pila(i)
c        IF(p.GE.2013)THEN
c           CALL build(p,pila(i+1),s,j)
c           WRITE(61,'(a,$)')' '//s(1:j)//' # '
c           i=i+2
c        ELSEIF(p.EQ.label)THEN
c           WRITE(61,'(a,i5,$)')' lab ',pila(i+1),' # '
c           i=i+2
c        ELSE
c           IF(p.EQ.add)WRITE(61,'(a,$)')' + '
c           IF(p.EQ.sub)WRITE(61,'(a,$)')' - '
c           IF(p.EQ.mul)WRITE(61,'(a,$)')' * '
c           IF(p.EQ.div)WRITE(61,'(a,$)')' / '
c           IF(p.EQ.pow)WRITE(61,'(a,$)')' ** '
c           IF(p.EQ.left)WRITE(61,'(a,$)')' ( '
c           IF(p.EQ.right)WRITE(61,'(a,$)')' ) '
c           IF(p.EQ.assign)WRITE(61,'(a,$)')' = '
c           IF(p.EQ.nline)WRITE(61,'(a,$)')' nl '
c           i=i+1
c        ENDIF
c     ENDDO
c     WRITE(61,*)
c     END
